module Parser where

import Char
import Exception
import AbsSynt
import Env

import Lex
import ParserMonad

-- Combinators for the BNF --
infixr 1 `symThenP`
infixl 2 `chkSymP`

pair x y = (x,y)

p   `chkSymP` sym = p           `chkP`  symbolP sym
sym `symThenP` p  = symbolP sym `thenP` p

star f = optionalP [] (plus f)
plus f = return (:) `apP` f `apP` optionalP [] plus_lem
  where plus_lem = "," `symThenP` plus f

wrapP l r p = l `symThenP` p `chkSymP` r

inParenP = wrapP "(" ")"
inBracketsP = wrapP "[" "]" 

parseNames :: Parse [Name]
parseNames = star parseName

parseIden :: Parse String
parseIden = idenP

parseName :: Parse Name
parseName  = parseIden

parseNs = manyP parseName

parseBind :: Parse a -> Parse b -> Parse (a,b)
parseBind f g = return pair `apP` f `apP` g

parseBinds :: Parse a -> Parse b -> Parse [(a,b)]
parseBinds f g = star (parseBind f g)

programP = return makeProg `apP` envP `apP` termP
           where makeProg = (curry Program)

envP = return makeEnv `apP` defsP
       where makeEnv = Env

defsP =  manyP defP

defP = return makeDef `apP` idenP `chkSymP` "=" `apP` termP `chkSymP` ";"
       where makeDef = (curry Def)

termP =  lambdaP
          `orP`
         ifP
          `orP`
         expr1

-- Abstraction 

lambdaP = return lamExp `chkSymP` "\\" `apP` parseNs `chkSymP` "->" `apP` termP
         where lamExp = flip $ foldr Lam

-- If sentence

ifP = return SIf `chkSymP` "if" `apP` expr1 
                 `chkSymP` "then" `apP` termP `chkSymP` "else" `apP` termP

-- Expressions

expr1 = return assembleOp `apP` expr2 `apP` expr1c
expr1c = return FoundOp `apP` (operP "||") `apP` expr1
         `orP`
         return NoOp
 
expr2 = return assembleOp `apP` expr3 `apP` expr2c
expr2c = return FoundOp `apP` (operP "&&") `apP` expr2
         `orP`
         return NoOp
 
expr3 = return assembleOp `apP` expr4 `apP` expr3c
expr3c = return FoundOp `apP` relopP `apP` expr4
         `orP`
         return NoOp
 
expr4 = return assembleOp `apP` expr5 `apP` expr4c
expr4c = return FoundOp `apP` (operP "+") `apP` expr4
         `orP` 
         return FoundOp `apP` (operP "-") `apP` expr5
         `orP`
         return NoOp

expr5 = return assembleOp `apP` expr6 `apP` expr5c
expr5c = return FoundOp `apP` (operP "*") `apP` expr5
         `orP` 
         return FoundOp `apP` (operP "/") `apP` expr6
         `orP`
         return NoOp
 
expr6 = appls

{- object `orP`  atomP -}

-- Application(s)

appls = return app `apP` someP term1P
        where app = foldl1 App

term1P = selOrUpdP 

-- Select or Update

selOrUpdP = return assembleUpD `apP` selectionP `apP` updateP

selectionP = 
     return proj `apP` term2P `apP` manyP ("." `symThenP` parseName)
     where
          proj t sels = foldl Sel t sels

updateP =  
        return FoundUpd `apP` ("<-"  `symThenP` methodP)
         `orP`
        return foundAssg `apP` (":="  `symThenP` termP)
         `orP`
        return NoArrowAssg
        where foundAssg o = FoundUpd (Met ("_",o)) -- "_" is a fake variable

term2P = objectP `orP` atomP

objectP = return mkobject `apP` (inBracketsP parseAssigs)
          where mkobject = Obj
                parseAssig = "=" `symThenP` methodP
                parseAssigs = parseBinds parseName parseAssig
               
methodP = 
   return mkmethod `apP` inParenP idenP `apP` termP
   where mkmethod x e = Met (x,e)       

atomP =
   factor
   `orP` 
   return id `apP` inParenP termP

factor = idenP >>=  \id -> return (if (all isDigit id) 
                                   then Num (read id)
                                   else if id == "True" then TT
                                        else if id == "False" then FF
                                             else Var id)

-- parser --

parseObj = parse termP . lexanal
parseDef = parse defP . lexanal
parseDefs = parse defsP . lexanal
parseProg = parse programP . lexanal

